# -*- coding: utf-8 -*-

import json
import uuid
from datetime import datetime
import apache_beam as beam
from beamism.utils import *


class PubSubLambda(beam.DoFn):
    def __init__(self,
                 func,
                 input_schema,
                 pipeline_name='',
                 mongo_uri=None,
                 mongo_db=None,
                 mongo_coll=None,
                 func_version=None):
        self.func = func
        self.input_schema = input_schema
        self.pipeline_name = pipeline_name
        self.mongo_uri = mongo_uri
        self.mongo_db = mongo_db
        self.mongo_coll = mongo_coll
        self.func_version = func_version

    def save(self, output_json):
        if self.mongo_db is not None and self.mongo_coll is not None:
            is_acknowledged = save_to_mongo(self.mongo_uri, self.mongo_db, self.mongo_coll, output_json)
            return is_acknowledged
        return False

    def process(self, msg_text):
        element_id = str(uuid.uuid4())
        start_time = datetime.now()

        input_json = parse_json_message(msg_text)
        if input_json is None:
            output_json = {
                "params": { "input_text": msg_text, "error": "parse error" },
                "element_id": element_id,
                "data": None,
                "succeeded": False,
                "message": "Cannot parse this message as json",
                "start_time": str(start_time),
                "end_time": str(datetime.now()),
                "processing_time": (datetime.now() - start_time).total_seconds(),
                "func_version": self.func_version
            }
            self.save(output_json)
            return pubsub_packing(output_json)

        if self.input_schema is not None:
            input_validation_err_message = validate_json_based_on_schema(input_json, self.input_schema)

            if input_validation_err_message is not None:
                output_json = {
                    "params": { "input_text": msg_text, "error": "validation error" },
                    "element_id": element_id,
                    "data": None,
                    "succeeded": False,
                    "message": input_validation_err_message,
                    "start_time": str(start_time),
                    "end_time": str(datetime.now()),
                    "processing_time": (datetime.now() - start_time).total_seconds(),
                    "func_version": self.func_version
                }
                self.save(output_json)
                return pubsub_packing(output_json)

        if not input_json["succeeded"] or input_json["data"] is None:
            output_jsons = {
                "params": input_json["data"],
                "sender_element_id": input_json["element_id"] if "element_id" in input_json else None,
                "element_id": element_id,
                "data": None,
                "succeeded": False,
                "message": input_json["message"],
                "start_time": str(start_time),
                "end_time": str(datetime.now()),
                "processing_time": (datetime.now() - start_time).total_seconds(),
                "func_version": self.func_version
            }
            self.save(output_json)
            return pubsub_packing(output_json)

        indata = input_json["data"]
        results, is_func_success, func_msg = self.func(indata)

        if type(results) != list:
            results = [{}] if results is None else [results]

        output_json_for_mongo = { 
            "params": input_json["data"],
            "sender_element_id": input_json["element_id"] if "element_id" in input_json else None,
            "element_id": element_id,
            "data": results,
            "succeeded": is_func_success,
            "message": func_msg,
            "start_time": str(start_time),
            "end_time": str(datetime.now()),
            "processing_time": (datetime.now() - start_time).total_seconds(),
            "func_version": self.func_version 
        }
        self.save(output_json_for_mongo)

        outdatas = [merge_jsons(indata, result) for result in results]
        output_jsons = [{ "params": input_json["data"],
                          "sender_element_id": input_json["element_id"] if "element_id" in input_json else None,
                          "element_id": element_id,
                          "data": outdata,
                          "succeeded": is_func_success,
                          "message": func_msg,
                          "start_time": str(start_time),
                          "end_time": str(datetime.now()),
                          "processing_time": (datetime.now() - start_time).total_seconds(),
                          "func_version": self.func_version } for outdata in outdatas]
        return pubsub_packing(output_jsons)



# -*- coding: utf-8 -*-


import json
import jsonschema
import copy
from pymongo import MongoClient


def save_to_mongo(mongo_uri, db_name, collection_name, data, ssl=True):
    # copy data since coll.insert_one function will generate new object in the data, and cannot serialize it
    cp_data = copy.copy(data)
    client = MongoClient(mongo_uri, ssl=ssl)
    db = client[db_name]
    coll = db[collection_name]
    insert_result = coll.insert_one(cp_data)
    return insert_result.acknowledged


def parse_json_message(json_as_str):
    try:
        input_json = json.loads(json_as_str)
        return input_json
    except:
        return None


def validate_json_based_on_schema(input_json, schema):
    try:
        jsonschema.validate(input_json, schema)
        return None
    except jsonschema.ValidationError as validation_error:
        return validation_error


def pubsub_packing(json_datas):
    '''
    How to Pack:
        convert msgs as list(json) => msgs as list[str(json style)]
    Note:
        When return string as message, pubsub treat it as list of chars. But, in many cases, we do not want such behaviour.
    '''
    if type(json_datas) != list:
        json_datas = [json_datas]
    return [json.dumps(json_data) for json_data in json_datas]


def merge_jsons(params_json, result_json):
    """
    Merge two jsons: the first one should be the parameters for the job, and the other should be the results of the job
    """
    # make sure keys not change
    if set(params_json.keys()).intersection(result_json.keys()):
        err_msg = "[BUG] When merge two jsons, some keys are modified"
        print(err_msg)

    merged_json = copy.copy(result_json)
    merged_json.update(params_json)
    return merged_json


# -*- coding: utf-8 -*-
#!/usr/bin/env python

from setuptools import setup, find_packages
import git


try:
    with open('README.md') as f:
        readme = f.read()
except IOError:
    readme = ''


def _requires_from_file(filename):
    return open(filename).read().splitlines()


try:
    current_commit_number = git.Repo(search_parent_directories=True).head.object.count() + 1
    zero_padding = '%03d' % current_commit_number
    version = '.'.join(list(zero_padding))
    with open('version.cache', 'w') as f:
        f.write(version)
except git.exc.InvalidGitRepositoryError:
    # Dataflow Deploy時、.gitフォルダはアップロードされないため、gitの情報をバージョンに利用できない。そのため、version.cacheからversion情報を取得する
    with open('version.cache', 'r') as f:
        version = f.read()


setup(
    name="beamism",
    version=version,
    packages=find_packages(),
    description='welcome to beamism',
    long_description=readme,
    url='https://github.com/kyu999/beamism',
    author='kyu999',
    author_email='kyukokkyou999@gmail.com',
    maintainer='kyu999',
    maintainer_email='kyukokkyou999@gmail.com',
    platforms='Linux, Darwin',
    zip_safe=False,
    include_package_data=True,
    install_requires=_requires_from_file('requirements.txt'),
    license="MIT",
    classifiers=[
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'License :: OSI Approved :: MIT License',
    ]
)
